/* Copyright (c) 2013 APDM
*
* gait_cycle.js
*
* Author: Jess VanDerwalker
*/

function gaitCycleChart() {
    // Physical layouts
    var margin = { right:20, top:20, left:60, bottom:60 };
    var height = 200;
    var width = 1000;

    // Scales and axies
    var xScale = d3.scale.linear();
    var yScale = d3.scale.ordinal()
        .domain(["Right", "Left"]);
    var xAxis = d3.svg.axis()
        .scale(xScale).orient("bottom");
    var yAxis = d3.svg.axis()
        .scale(yScale).orient("left");

    function gaitChart(selection) {
        selection.each(function(metrics) {

            var gaitMetrics = processGaitCycleMetrics(metrics);

            // Update scales and axis functions
            xScale.range([0, width - margin.left - margin.right]).domain([0,100]);
            yScale.rangeRoundBands([height - margin.top - margin.bottom, 0]);

            var chartSvg = d3.select(this).selectAll("svg").data([gaitMetrics]);
            var chart = chartSvg.enter().append("svg").append("g");

            chart.append("g").attr("class", "gait_rects");

            // Update chart components
            chartSvg.attr("width", width).attr("height", height);

            chart = chartSvg.select("g")
                .attr("transform", "translate(" + margin.left + "," + margin.top + ")");
            
            chart.select(".gait_rects").selectAll("rect")
                .data(gaitMetrics.gaitRects)
                .enter()
                .append("rect")
                .each( function(d) {
                    var rect = d3.select(this);
                    var y = "Left";
                    if (d.type.substring(d.type.length - 1, d.type.length) === 'r')
                        y = "Right";
                    
                    rect.attr("x", xScale(d.start))
                        .attr("y", yScale(y))
                        .attr("height", (height - margin.top - margin.bottom) / 2)
                        .attr("width", xScale(d.end - d.start))
                        .attr("class", d.type + " gait_track_rect");
                });

            // Rectangle for showing the double support area
            var rects = chart.select(".gait_rects");
            rects.append("rect")
                .attr("class", "dbl_support")
                .attr("x", xScale(gaitMetrics.doubleRange.start))
                .attr("y", (height - margin.top - margin.bottom) / 4)
                .attr("height", (height - margin.top - margin.bottom) / 2)
                .attr("width", xScale(gaitMetrics.doubleRange.end -
                                      gaitMetrics.doubleRange.start));
                     
            // Axis for the chart
            chart.append("g").attr("class", "x axis")
                .attr("transform", "translate(0," +
                      (height - margin.top - margin.bottom) + ")")
                .call(xAxis);
            chart.append("g").attr("class", "y axis")
                .call(yAxis);

        });
    }

    function processGaitCycleMetrics(metrics) {
        metrics = d3.nest().key(function(d) { return d.metricName; })
            .map(metrics, d3.map);

        // Need to determine the start and end points for each part of
        // the gait cycle. We start with the left swing.
        var gaitRects = [];
        var swingLRect = { name: metrics["Swing L"][0].metricName,
                           type: (metrics["Swing L"][0].metricName).replace(/ /g, "_")
                           .toLowerCase(),
                           start: 0,
                           end: metrics["Swing L"][0].mean,
                         };
        gaitRects.push(swingLRect);

        var stanceLRect = { name: metrics["Stance L"][0].metricName,
                            type: (metrics["Stance L"][0].metricName).replace(/ /g, "_")
                            .toLowerCase(),
                            start: swingLRect.end,
                            end: metrics["Swing L"][0].mean + metrics["Stance L"][0].mean,
                          };
        gaitRects.push(stanceLRect);

        var stanceRRect = { name: metrics["Stance R"][0].metricName,
                            type: (metrics["Stance R"][0].metricName).replace(/ /g, "_")
                            .toLowerCase(),
                            start: 0,
                            end: metrics["Stance R"][0].mean,
                          };
        gaitRects.push(stanceRRect);

        var swingRRect = { name: metrics["Swing R"][0].metricName,
                           type: (metrics["Swing R"][0].metricName).replace(/ /g, "_")
                           .toLowerCase(),
                           start: metrics["Stance R"][0].mean,
                           end: metrics["Stance R"][0].mean + metrics["Swing R"][0].mean,
                         };
        gaitRects.push(swingRRect);

        // Return the bounds for the double support rect as well
        return { gaitRects: gaitRects,
                 doubleRange: {start: swingLRect.end,
                               end: swingRRect.start}
               };
    }

    // Physical configuration accessors
    gaitChart.width = function(d) {
        if (!arguments.length) return width;
        width = d;
        return gaitChart;
    };
    gaitChart.height = function(d) {
        if (!arguments.length) return height;
        height = d;
        return gaitChart;
    };
    gaitChart.margin = function(d) {
        if (!arguments.length) return margin;
        margin = d;
        return gaitChart;
    };

    return gaitChart;
}
