/* Copyright (c) 2013 APDM
*
* single_sensor_plot.js
*
* Author: Jess VanDerwalker
*/

function singleSensorPlot () {
    var height = 400;
    var width = 800;
    var margin = { top: 20, right: 20, bottom: 20, left: 20 };

    // Should we try and plot a set of secondary lines.
    var showSecondary = false;

    // Should we try and plot events
    var showEvents = false;

    // Should the time be displayed as relative (e.g. start time = 0)
    var relativeTime = false;

    // Labels for the vertical and horizontal axises
    var vAxisLabel = "";
    var hAxisLabel = "";

    // Flags controlling if axis ticks are displayed
    var drawHorzTicks = true;

    // We have four lines and the time for primary and secondary
    var time1_Value = function (d) { return d[0]; };
    var x1_Value = function (d) { return d[1]; };
    var y1_Value = function (d) { return d[2]; };
    var z1_Value = function (d) { return d[3]; };
    var scalar1_Value = function (d) { return d[4]; };

    var time2_Value = function (d) { return d[5]; };
    var x2_Value = function (d) { return d[6]; };
    var y2_Value = function (d) { return d[7]; };
    var z2_Value = function (d) { return d[8]; };
    var scalar2_Value = function (d) { return d[9]; };

    // Accessors for the event data
    var startTime = function (d) { return d[0]; };
    var endTime = function (d) { return d[1]; };
    var evType = function (d) { return d[2]; };
    var evID = function (d) { return d[3]; };

    // Both the primary and secondary will share the same scales. 
    var timeScale = d3.time.scale();
    var vertScale = d3.scale.linear();

    // Custom time format for our time ticks
    var xTickFormat = timeFormat([
        [d3.time.format.utc("%Y"), function() { return true; }],
        [d3.time.format.utc("%B"), function(d) { return d.getMonth(); }],
        [d3.time.format.utc("%b %d"), function(d) { return d.getDate() !== 1; }],
        [d3.time.format.utc("%a %d"), function(d) { return d.getDay() && d.getDate() !== 1; }],
        [d3.time.format.utc("%H"), function(d) { return d.getHours(); }],
        [d3.time.format.utc("%H:%M"), function(d) { return d.getMinutes(); }],
        [d3.time.format.utc(":%S"), function(d) { return d.getSeconds(); }],
        [d3.time.format.utc(".%L"), function(d) { return d.getMilliseconds(); }]
    ]);
    function timeFormat(formats) {
        return function(date) {
            var i = formats.length - 1, f = formats[i];
            while (!f[1](date)) f = formats[--i];
            return f[0](date);
        };
    }

    // Axis ticks and grid
    var timeAxisTicks = d3.svg.axis()
        .orient("bottom").ticks(5);
    var timeAxisGrid = d3.svg.axis().scale(timeScale)
        .orient("bottom").tickFormat("").ticks(5);
    var vertAxis = d3.svg.axis().scale(vertScale).orient("left").ticks(5);
    var vertAxisGrid = d3.svg.axis().scale(vertScale).
        orient("left").tickFormat("").ticks(5);

    // Line functions for each of the graphs
    // Primary lines
    var x1_Line = d3.svg.line().x(TIME_1).y(X_1);
    var y1_Line = d3.svg.line().x(TIME_1).y(Y_1);
    var z1_Line = d3.svg.line().x(TIME_1).y(Z_1);
    var scalar1_Line = d3.svg.line().x(TIME_1).y(SCALAR_1);

    // Secondary lines
    var x2_Line = d3.svg.line().x(TIME_2).y(X_2);
    var y2_Line = d3.svg.line().x(TIME_2).y(Y_2);
    var z2_Line = d3.svg.line().x(TIME_2).y(Z_2);
    var scalar2_Line = d3.svg.line().x(TIME_2).y(SCALAR_2);

    // Scaling and transform for the plot
    var plotScale = 1;
    var plotTranslate = [0,0];

    function sensorPlot(selection) {
        selection.each(function (plotData) {
            // Determine if we have event data
            var events;
            if (plotData.length === 2) {
                events = plotData[1].map(function(d, i) {
                    return [ startTime.call(plotData[1], d, i),
                             endTime.call(plotData[1], d, i),
                             evType.call(plotData[1], d, i),
                             evID.call(plotData[1], d, i) ];
                });
            }
            // Convert the data into a nice matrix, padding the arrays
            // if we aren't plotting every line.
            var data = plotData[0].map(function(d, i) {
                var entry = [];
                entry.push(time1_Value.call(plotData[0], d, i));
                entry.push(x1_Value ? x1_Value.call(plotData[0], d, i) : 0);
                entry.push(y1_Value ? y1_Value.call(plotData[0], d, i) : 0);
                entry.push(z1_Value ? z1_Value.call(plotData[0], d, i) : 0);
                entry.push(scalar1_Value ? scalar1_Value.call(plotData[0], d, i) : 0);

                // Add secondary plotData[0], if set
                if (showSecondary) {
                    entry.push(time2_Value.call(plotData[0], d, i));
                    entry.push(x2_Value ? x2_Value.call(plotData[0], d, i) : 0);
                    entry.push(y2_Value ? y2_Value.call(plotData[0], d, i) : 0);
                    entry.push(z2_Value ? z2_Value.call(plotData[0], d, i) : 0);
                    entry.push(scalar2_Value ? scalar2_Value.call(plotData[0], d, i) : 0);
                }
                return entry;
            });
            
            var startXTime = data[0][0];

            // If plot is using relative time, remove the start time from every time value.
            if (relativeTime) {
                timeScale = d3.time.scale.utc();
                timeAxisTicks = d3.svg.axis().scale(timeScale)
                    .orient("bottom").ticks(5).tickFormat(xTickFormat);
                data = data.map(function(d) {
                    d[0] = d[0] - startXTime;
                    return d;
                });
                if (showSecondary) {
                    data = data.map(function(d) {
                        d[5] = d[5] - startXTime;
                        return d;
                    });
                }
                if (showEvents && events) {
                    events = events.map(function(d) {
                        d[0] = d[0] - startXTime;
                        d[1] = d[1] - startXTime;
                        return d;
                    });
                }
            } else {
                timeScale = d3.time.scale();
                timeAxisTicks = d3.svg.axis().scale(timeScale)
                    .orient("bottom").ticks(5);

            }
            timeAxisGrid = d3.svg.axis().scale(timeScale)
                .orient("bottom").tickFormat("").ticks(5);
            vertAxis = d3.svg.axis().scale(vertScale).orient("left").ticks(5);
            vertAxisGrid = d3.svg.axis().scale(vertScale)
                .orient("left").tickFormat("").ticks(5);

            // Time scale based only on primary data
            timeScale.domain(d3.extent(data, function(d) { return d[0]; }))
                .range([0, width - margin.left - margin.right]);

            // Find the extents for each line
            var allYValues = data.map(function(d) { return d[1]; })
                .concat(data.map(function(d) { return d[2]; }),
                        data.map(function(d) { return d[3]; }));
            if (scalar1_Value) {
                allYValues = allYValues.concat(data.map(function(d) { return d[4]; }));
            }

            if (showSecondary) {
                allYValues = allYValues.concat(data.map(function(d) { return d[6]; }),
                                              data.map(function(d) { return d[7]; }),
                                              data.map(function(d) { return d[8]; }));
                if (scalar2_Value) {
                    allYValues = allYValues.concat(data.map(function(d) { return d[9]; }));
                }
            }

            vertScale.domain(d3.extent(allYValues))
                .range([height - margin.top - margin.bottom, 0]);
            
            // Limit panning to range of data
            plotTranslate[0] = Math.min(plotTranslate[0], 0);
            plotTranslate[0] = Math.max(plotTranslate[0],
                                        (width - margin.left - margin.right) -
                                        timeScale.range()[1] * plotScale);

            // Update the scales with the current translates and scales
            // We only allow zoom and pan in the x-axis.
            var timeScaleCopy = timeScale.copy();
            timeScale.domain(timeScaleCopy.range().map(function(x) {
                return (x - plotTranslate[0]) / plotScale;
            }).map(timeScaleCopy.invert));

            // Create the elements to hold the plots
            var plotSvg = d3.select(this).selectAll("svg").data([data]);
            var plotG = plotSvg.enter().append("svg").append("g");

            // We have to do all the "appends" before any selects with
            // plotG, otherwise it creates new elements each time
            // through.
            
            // Create a clip path to limit plot when zoomed
            plotG.append("defs").append("clipPath")
                .attr("id", "clip").append("rect")
                .attr("x", 0).attr("y", 0)
                .attr("width", width - margin.left - margin.right)
                .attr("height", height - margin.top - margin.bottom);

            // "g" ontainers for the axises
            plotG.append("g").attr("class", "x axis");
            plotG.append("g").attr("class", "y axis");
            plotG.append("g").attr("class", "grid vert");
            plotG.append("g").attr("class", "grid horz");

            // Labels for the axises
            plotG.append("text")
                .attr("class", "y label");
            plotG.append("text")
                .attr("class", "x label");

            // Only do the zero x label if we are also showing the other tick values.
            if (drawHorzTicks) {
                plotG.append("rect")
                    .attr("class", "zero_x_tick_mask");
                plotG.append("text")
                    .attr("class", "zero_x_tick label");
            }

            // Create primary lines for the plots
            plotG.append("path").attr("class", "line_x primary_lines")
                .attr("clip-path", "url(#clip)");
            plotG.append("path").attr("class", "line_y primary_lines")
                .attr("clip-path", "url(#clip)");
            plotG.append("path").attr("class", "line_z primary_lines")
                .attr("clip-path", "url(#clip)");
            plotG.append("path").attr("class", "line_scalar primary_lines")
                .attr("clip-path", "url(#clip)");

            // Secondary lines, if we should show them
            if (showSecondary) {
                plotG.append("path").attr("class", "line_x secondary_lines")
                    .attr("clip-path", "url(#clip)");
                plotG.append("path").attr("class", "line_y secondary_lines")
                    .attr("clip-path", "url(#clip)");
                plotG.append("path").attr("class", "line_z secondary_lines")
                    .attr("clip-path", "url(#clip)");
                plotG.append("path").attr("class", "line_scalar secondary_lines")
                    .attr("clip-path", "url(#clip)");
            }

            if (showEvents) {
                plotG.append("g").attr("class", "events_g");

                plotG.select(".events_g").selectAll("rect")
                    .data(events, function(d) { return d; })
                    .enter()
                    .append("rect")
                    .attr("class", function(d) {
                        return "event_rect " + EVT_TYPE(d); })
                    .attr("id", function(d, i) { return EVT_ID(d) + "-" + i; })
                    .attr("x", function(d) { return EVT_START(d); })
                    .attr("y", "0")
                    .attr("width", function(d) { return EVT_END(d) - EVT_START(d); })
                    .attr("height", height - margin.bottom - margin.top)
                    .attr("clip-path", "url(#clip)");

                // TODO: Seems that there should be a cleaner way to
                // so this using D3 enter() and update()
                events.forEach( function(d,i) {
                    var rect = plotSvg.select(".events_g")
                        .select("#" + EVT_ID(d) + "-" + i);
                    rect.attr("x", EVT_START(d))
                        .attr("y", "0")
                        .attr("width", EVT_END(d) - EVT_START(d))
                        .attr("height", height - margin.bottom - margin.top);
                });
            }

            // Update the components of the plot
            plotSvg.attr("width", width).attr("height", height);

            // This handles updates to inner dimentions
            plotG = plotSvg.select("g")
                .attr("transform", "translate(" + (margin.left) +
                      "," + (margin.top) + ")");

            // Update our clip path
            plotG.select("#clip").select("rect")
                .attr("x", 0).attr("y", 0)
                .attr("width", width - margin.left - margin.right)
                .attr("height", height- margin.top - margin.bottom);

            // Update the axises and the grid, setting grid size to given margins
            timeAxisGrid.tickSize(height - margin.top - margin.bottom, 0, 0);
            vertAxisGrid.tickSize(-width + margin.left + margin.right, 0, 0);
            if (!drawHorzTicks)
                timeAxisTicks.tickFormat("");
            plotG.select(".x.axis")
                .attr("transform", "translate(0," + vertScale.range()[0] + ")")
                .call(timeAxisTicks);
            plotG.select(".grid.vert").attr("transform", "translate(0,0)")
                .call(vertAxisGrid);
            plotG.select(".grid.horz").attr("transform", "translate(0,0)")
                .call(timeAxisGrid);
            plotG.select(".y.axis")
                .attr("transform", "translate(" + timeScale.range()[0] + ",0)")
                .call(vertAxis);

            // Update axis labels
            var tickPad = timeAxisTicks.tickPadding();
            var tickSize = timeAxisTicks.tickSize();

            plotG.select(".y.label")
                .attr("text-anchor", "middle")
                .attr("x", (margin.top + margin.bottom - height)/2)
                .attr("y", -35)
                .attr("transform", "rotate(-90)")
                .text(vAxisLabel);

            var xLabel = plotG.select(".x.label")
                .attr("text-anchor", "middle")
                .attr("x", (width - margin.left - margin.right)/2)
                .attr("y", (height - margin.top - margin.bottom))
                .text(hAxisLabel);

            var box = xLabel.node().getBBox();
            var hOffset = (height - margin.top - margin.bottom) +
                tickPad + tickSize + box.height + (box.height / 2);
            xLabel.attr("y", hOffset);

            if (drawHorzTicks) {
                // The zero axis tick label needs to update with the zoom
                // and pan of the plot.
                var zeroTickLabel = plotG.select(".zero_x_tick.label")
                    .attr("text-anchor", "middle")
                    .attr("x", 0)
                    .attr("y", height - margin.top - margin.bottom + tickPad + tickSize)
                    .text(function() {
                        var format;
                        if (relativeTime) {
                            format = d3.time.format.utc("%H:%M:%S");
                        } else {
                            format = d3.time.format("%H:%M:%S");
                        }
                        return format(timeScale.invert(0));
                    });
                box = zeroTickLabel.node().getBBox();
                zeroTickLabel.attr("y", height - margin.top - margin.bottom +
                                   tickPad + tickSize + (box.height / 2));
                plotG.select(".zero_x_tick_mask")
                    .attr("x", box.x)
                    .attr("y", box.y + (box.height / 2))
                    .attr("height", box.height)
                    .attr("width", box.width);
            }
            
            // Update line paths
            if (showSecondary) {
                // Update secondary paths
                plotG.select(".line_x.secondary_lines")
                    .attr("d", (x2_Value ? x2_Line : "M0 0"));
                plotG.select(".line_y.secondary_lines")
                    .attr("d", (y2_Value ? y2_Line : "M0 0"));
                plotG.select(".line_z.secondary_lines")
                    .attr("d", (z2_Value ? z2_Line : "M0 0"));
                plotG.select(".line_scalar.secondary_lines")
                    .attr("d", (scalar2_Value ? scalar2_Line : "M0 0"));
            }

            // Update primary paths
            plotG.select(".line_x.primary_lines")
                .attr("d", (x1_Value ? x1_Line : "M0 0"));
            plotG.select(".line_y.primary_lines")
                .attr("d", (y1_Value ? y1_Line : "M0 0"));
            plotG.select(".line_z.primary_lines")
                .attr("d", (z1_Value ? z1_Line : "M0 0"));
            plotG.select(".line_scalar.primary_lines")
                .attr("d", (scalar1_Value ? scalar1_Line : "M0 0"));
        });
    }

    // Primary accessors
    function TIME_1(d) {
        return timeScale(d[0]);
    }
    function X_1(d) {
        return vertScale(d[1]);
    }
    function Y_1(d) {
        return vertScale(d[2]);
    }
    function Z_1(d) {
        return vertScale(d[3]);
    }
    function SCALAR_1(d) {
        return vertScale(d[4]);
    }

    // Secondary accessors
    function TIME_2(d) {
        return timeScale(d[5]);
    }
    function X_2(d) {
        return vertScale(d[6]);
    }
    function Y_2(d) {
        return vertScale(d[7]);
    }
    function Z_2(d) {
        return vertScale(d[8]);
    }
    function SCALAR_2(d) {
        return vertScale(d[9]);
    }

    // Event accessors
    function EVT_START(d) {
        return timeScale(d[0]);
    }
    function EVT_END(d) {
        return timeScale(d[1]);
    }
    function EVT_TYPE(d) {
        return d[2];
    }
    function EVT_ID(d) {
        return d[3];
    }


    sensorPlot.width = function(d) {
        if (!arguments.length) return width;
        width = d;
        return sensorPlot;
    };
    sensorPlot.height = function(d) {
        if (!arguments.length) return height;
        height = d;
        return sensorPlot;
    };

    // Primary
    sensorPlot.time1 = function(d) {
        if (!arguments.length) return time1_Value;
        time1_Value = d;
        return sensorPlot;
    };
    sensorPlot.x1 = function(d) {
        if (!arguments.length) return x1_Value;
        x1_Value = d;
        return sensorPlot;
    };
    sensorPlot.y1 = function(d) {
        if (!arguments.length) return y1_Value;
        y1_Value = d;
        return sensorPlot;
    };
    sensorPlot.z1 = function(d) {
        if (!arguments.length) return z1_Value;
        z1_Value = d;
        return sensorPlot;
    };
    sensorPlot.scalar1 = function(d) {
        if (!arguments.length) return scalar1_Value;
        scalar1_Value = d;
        return sensorPlot;
    };

    // Secondary functions
    sensorPlot.time2 = function(d) {
        if (!arguments.length) return time2_Value;
        time2_Value = d;
        return sensorPlot;
    };
    sensorPlot.x2 = function(d) {
        if (!arguments.length) return x2_Value;
        x2_Value = d;
        return sensorPlot;
    };
    sensorPlot.y2 = function(d) {
        if (!arguments.length) return y2_Value;
        y2_Value = d;
        return sensorPlot;
    };
    sensorPlot.z2 = function(d) {
        if (!arguments.length) return z2_Value;
        z2_Value = d;
        return sensorPlot;
    };
    sensorPlot.scalar2 = function(d) {
        if (!arguments.length) return scalar2_Value;
        scalar2_Value = d;
        return sensorPlot;
    };

    // Event functions
    sensorPlot.eventStart = function (d) {
        if (!arguments.length) return startTime;
        startTime = d;
        return sensorPlot;
    };
    sensorPlot.eventEnd = function(d) {
        if (!arguments.length) return endTime;
        endTime = d;
        return sensorPlot;
    };
    sensorPlot.eventType = function(d) {
        if (!arguments.length) return evType;
        evType = d;
        return sensorPlot;
    };
    sensorPlot.eventID = function(d) {
        if (!arguments.length) return evID;
        evID = d;
        return sensorPlot;
    };


    // Scale functions
    sensorPlot.xScale = function(d) {
        if (!arguments.length) return timeScale;
        timeScale = d;
        return sensorPlot;
    };
    sensorPlot.yScale = function(d) {
        if (!arguments.length) return vertScale;
        vertScale = d;
        return sensorPlot;
    };


    // Configuration settings
    sensorPlot.margin = function(d) {
        if (!arguments.length) return margin;
        margin = d;
        return sensorPlot;
    };
    sensorPlot.scale = function(d) {
        if (!arguments.length) return plotScale;
        plotScale = d;
        return sensorPlot;
    };
    sensorPlot.translate = function (d) {
        if (!arguments.length) return plotTranslate;
        plotTranslate[0] = d[0];
        plotTranslate[1] = d[1];
        return sensorPlot;
    };

    sensorPlot.vertLabel = function(d) {
        if (!arguments.length) return vAxisLabel;
        vAxisLabel = d;
        return sensorPlot;
    };

    sensorPlot.horzLabel = function(d) {
        if (!arguments.length) return hAxisLabel;
        hAxisLabel = d;
        return sensorPlot;
    };

    sensorPlot.displayRelativeTime = function(d) {
        if (!arguments.length) return relativeTime;
        relativeTime = d;
        return sensorPlot;
    };

    sensorPlot.showHorzTicks = function(d) {
        if (!arguments.length) return drawHorzTicks;
        drawHorzTicks = d;
        return sensorPlot;
    };

    sensorPlot.plotSecondary = function (d) {
        if (!arguments.length) return showSecondary;
        showSecondary = d;
        return sensorPlot;
    };

    sensorPlot.plotEvents = function(d) {
        if (!arguments.length) return showEvents;
        showEvents = d;
        return sensorPlot;
    };

    return sensorPlot;
}
